from typing import List
from stellar_sdk import Network


DATABASE_URL: str = 'sqlite:///db.sqlite3'
ASYNC_DATABASE_URL: str = 'sqlite+aiosqlite:///db.sqlite3'

JWT_SECRET: str = "@)wqgb)3&e6k&(l8hfm(3wt8=*_x$w@vc$4)&nbih-&2eg9dlh"
SIGNING_SECRET: str = "SD3ME2YQNWQYBKYX7KNMX5C42WTWMZRZD7DH72K63B56G636AYBQH7YY"
DISTRIBUTION_SECRET: str = "SC5N4HUKY55KO6IETCP4TNJPS4WXCQFRLDCQY5PPAMKZP24BXKMF4ZEU"

HOST_URL: str = "http://localhost"
HOME_DOMAINS: List[str] = ["localhost"]
HORIZON_URL: str = "https://horizon-testnet.stellar.org"
NETWORK_PASSPHRASE: str = Network.TESTNET_NETWORK_PASSPHRASE



import uuid
import asyncio
from decimal import Decimal
from datetime import datetime, timezone
from typing import List, Callable, Any, Optional, Union, Dict
from fastapi import FastAPI
from pydantic import Field
from databases import Database
from fastapi import Depends, Request
from overrides import overrides
from abc import ABC, abstractmethod
import fawaris

from fawaris_fastapi import tables
from fawaris_fastapi import settings
from fawaris_fastapi.utils import (
    authenticate,
    detect_and_get_request_data,
    fetch_account,
    is_pending_trust,
    stellar_create_claimable_balance,
    stellar_send_payment,
    get_claimable_balance_id,
    validate_request_data,
)

class Sep24Transaction(fawaris.Sep24Transaction):
    asset: fawaris.Asset = Field(..., exclude=True)
    paging_token: Optional[str] = Field(None, exclude=True)
    claimable_balance_supported: Optional[str] = Field(None, exclude=True)
    stellar_transaction_response: Optional[str] = Field(None, exclude=True)

    @staticmethod
    def from_db_row(row, *, assets: Dict[str, fawaris.Asset]):
        row = dict(row)
        asset_code = row.pop("asset_code")
        asset_issuer = row.pop("asset_issuer")
        transaction = Sep24Transaction(
            **{
                **row,
                "asset": assets[asset_code],
            }
        )
        return transaction

    @staticmethod
    def to_db_values(transaction):
        values = {
            "asset_code": transaction.asset.code,
            "asset_issuer": transaction.asset.issuer,
            "paging_token": transaction.paging_token,
            "claimable_balance_supported": transaction.claimable_balance_supported,
            "stellar_transaction_response": transaction.stellar_transaction_response,
        }
        values.update(transaction.dict())
        return values


class Sep24(fawaris.Sep24):
    database: Database
    sep10_jwt_secret: str
    distribution_secret: str
    horizon_url: str
    network_passphrase: str
    assets: Dict[str, fawaris.Asset]
    log: Callable[[str], Any]

    def __init__(
        self,
        database: Database,
        sep10_jwt_secret: str,
        distribution_secret: str,
        horizon_url: str,
        network_passphrase: str,
        assets: Dict[str, fawaris.Asset],
        log: Optional[Callable[[str], Any]] = lambda msg: None,
    ):
        self.database = database
        self.sep10_jwt_secret = sep10_jwt_secret
        self.distribution_secret = distribution_secret
        self.horizon_url = horizon_url
        self.network_passphrase = network_passphrase
        self.assets = assets
        self.log = log

    @classmethod
    def new_instance(cls, database: Database):
        return cls(
            database=database,
            sep10_jwt_secret=settings.JWT_SECRET,
            distribution_secret=settings.DISTRIBUTION_SECRET,
            horizon_url=settings.HORIZON_URL,
            network_passphrase=settings.NETWORK_PASSPHRASE,
            assets={
                "USDC": fawaris.Asset(
                    code="USDC",
                    issuer="GCCGEMWASPXY4HWAJTH2BDYKYMLVOAH4K657IDLM5ZE7DGJJCY5EY53J",
                    decimal_places=7,
                )
            },
        )

    def register_routes(self, app: FastAPI):
        @app.post("/sep24/transactions/deposit/interactive")
        async def http_post_transactions_deposit_interactive(
            request=Depends(authenticate(self.sep10_jwt_secret)),
        ):
            data = await detect_and_get_request_data(request)
            req = validate_request_data(data, fawaris.Sep24DepositPostRequest)
            return await self.http_post_transactions_deposit_interactive(
                req, request.token
            )

        @app.post("/sep24/transactions/withdraw/interactive")
        async def http_post_transactions_withdraw_interactive(
            request=Depends(authenticate(self.sep10_jwt_secret)),
        ):
            data = await detect_and_get_request_data(request)
            req = validate_request_data(data, fawaris.Sep24WithdrawPostRequest)
            return await self.http_post_transactions_withdraw_interactive(
                req, request.token
            )

        @app.get("/sep24/info")
        async def http_get_info(request: Request):
            req = validate_request_data(
                dict(request.query_params), fawaris.Sep24InfoRequest
            )
            return await self.http_get_info(req)

        @app.get("/sep24/transactions")
        async def http_get_transactions(
            request=Depends(authenticate(self.sep10_jwt_secret)),
        ):
            req = validate_request_data(
                dict(request.query_params), fawaris.Sep24TransactionsGetRequest
            )
            return await self.http_get_transactions(req, request.token)

        @app.get("/sep24/transaction")
        async def http_get_transaction(
            request=Depends(authenticate(self.sep10_jwt_secret)),
        ):
            req = validate_request_data(
                dict(request.query_params), fawaris.Sep24TransactionGetRequest
            )
            return await self.http_get_transaction(req, request.token)

    @overrides
    async def http_get_fee(
        self,
        request: fawaris.Sep24FeeRequest,
        token: Optional[fawaris.Sep10Token] = None,
    ) -> fawaris.Sep24FeeResponse:
        raise NotImplementedError()

    @overrides
    async def http_get_transactions(
        self, request: fawaris.Sep24TransactionsGetRequest, token: fawaris.Sep10Token
    ) -> fawaris.Sep24TransactionsGetResponse:
        query = tables.sep24_transactions.select()
        if request.asset_code is not None:
            query = query.where(
                tables.sep24_transactions.c.asset_code == request.asset_code
            )
        transactions = []
        async for row in self.database.iterate(query):
            transactions.append(Sep24Transaction.from_db_row(row, assets=self.assets))
        return transactions

    @overrides
    async def http_get_transaction(
        self, request: fawaris.Sep24TransactionGetRequest, token: fawaris.Sep10Token
    ) -> fawaris.Sep24TransactionGetResponse:
        raise NotImplementedError()

    @overrides
    async def create_transaction(
        self,
        request: Union[
            fawaris.Sep24DepositPostRequest, fawaris.Sep24WithdrawPostRequest
        ],
        token: fawaris.Sep10Token,
    ) -> Sep24Transaction:
        if isinstance(request, fawaris.Sep24DepositPostRequest):
            kind = "deposit"
        if isinstance(request, fawaris.Sep24WithdrawPostRequest):
            kind = "withdrawal"
        query = tables.sep24_transactions.insert()
        transaction = Sep24Transaction(
            id=str(uuid.uuid4()),  # TODO make sure it's unique
            kind=kind,
            status="pending_user_transfer_start",
            asset=self.assets[request.asset_code],
            to_address=request.account,
            amount_out="1234.56",
            external_extra="test",
        )
        values = Sep24Transaction.to_db_values(transaction)
        await self.database.execute(query=query, values=values)
        return transaction

    @overrides
    async def get_interactive_url(
        self,
        request: Union[
            fawaris.Sep24DepositPostRequest, fawaris.Sep24WithdrawPostRequest
        ],
        token: fawaris.Sep10Token,
        tx: fawaris.Sep24Transaction,
    ) -> str:
        return "https://sep24.mydomain.com"

    @overrides
    async def get_transactions(
        self,
        kind=None,
        status=None,
        memo=None,
        withdraw_anchor_account=None,
    ) -> List[fawaris.Sep24Transaction]:
        query = tables.sep24_transactions.select()
        if kind is not None:
            query = query.where(tables.sep24_transactions.c.kind == kind)
        if status is not None:
            query = query.where(tables.sep24_transactions.c.status == status)
        if memo is not None:
            query = query.where(tables.sep24_transactions.c.memo == memo)
        if withdraw_anchor_account is not None:
            query = query.where(
                tables.sep24_transactions.c.withdraw_anchor_account
                == withdraw_anchor_account
            )
        transactions = []
        async for row in self.database.iterate(query):
            transactions.append(Sep24Transaction.from_db_row(row, assets=self.assets))
        return transactions

    @overrides
    async def get_transaction_asset(
        self, transaction: fawaris.Sep24Transaction
    ) -> fawaris.Asset:
        return transaction.asset

    @overrides
    async def process_withdrawal_received(
        self,
        transaction: fawaris.Sep24Transaction,
        amount_received: str,
        from_address: str,
        horizon_response: Dict,
    ):
        if Decimal(amount_received) != Decimal(transaction.amount_in):
            await self.log_transaction_message(
                transaction,
                (
                    "Expected withdrawal amount_in={} differs from "
                    "received (stellar_transaction_id={}) withdrawal "
                    "amount_in={}".format(
                        transaction.amount_in,
                        horizon_response["id"],
                        amount_received,
                    )
                ),
            )
            await self.update_transactions(
                [transaction],
                status="error",
                stellar_transaction_id=horizon_response["id"],
                paging_token=horizon_response["paging_token"],
            )
        else:
            await self.log_transaction_message(
                transaction,
                "Withdrawal received with correct amount (stellar_transaction_id={})".format(
                    horizon_response["id"]
                ),
            )
            await self.update_transactions(
                [transaction],
                status="pending_anchor",
                from_address=from_address,
                stellar_transaction_id=horizon_response["id"],
                paging_token=horizon_response["paging_token"],
            )

    @overrides
    async def update_transactions(
        self, transactions: List[fawaris.Sep24Transaction], **values
    ):
        ids = [transaction.id for transaction in transactions]
        query = (
            tables.sep24_transactions.update()
            .where(tables.sep24_transactions.c.id.in_(ids))
            .values(dict(values))
        )
        coroutines = []
        for transaction in transactions:
            coroutines.append(
                self.log_transaction_message(transaction, f"Updating values: {values}")
            )
        await asyncio.gather(*coroutines)
        await self.database.execute(query)

    @overrides
    async def send_deposit(self, deposit: fawaris.Sep24Transaction) -> None:
        payment_args = dict(
            source_secret=self.distribution_secret,
            horizon_url=self.horizon_url,
            network_passphrase=self.network_passphrase,
            destination_account=deposit.to_address,
            amount=deposit.amount_out,
            asset_code=deposit.asset.code,
            asset_issuer=deposit.asset.issuer,
            memo=deposit.deposit_memo,
            memo_type=deposit.deposit_memo_type,
        )
        if deposit.claimable_balance_supported:
            account_dict = await fetch_account(self.horizon_url, deposit.to_address)
            if is_pending_trust(account_dict, deposit.asset.code, deposit.asset.issuer):
                await self.log_transaction_message(deposit, "Creating claimable balance")
                tx = await stellar_create_claimable_balance(
                    source_secret=self.distribution_secret,
                    horizon_url=self.horizon_url,
                    network_passphrase=self.network_passphrase,
                    destination_account=deposit.to_address,
                    amount=deposit.amount_out,
                    asset_code=deposit.asset.code,
                    asset_issuer=deposit.asset.issuer,
                    memo=deposit.deposit_memo,
                    memo_type=deposit.deposit_memo_type,
                )
                await self.update_transactions(
                    [deposit],
                    status="completed",
                    stellar_transaction_id=tx["transaction_hash"],
                    claimable_balance_id=tx[get_claimable_balance_id(tx)],
                )
                return None

        await self.log_transaction_message(
            deposit,
            "Sending stellar payment",
        )
        tx = await stellar_send_payment(**payment_args)
        await self.update_transactions(
            [deposit],
            status="completed",
            stellar_transaction_id=tx["transaction_hash"],
        )
        return None

    @overrides
    async def get_withdraw_anchor_account_cursor(self, account: str) -> Optional[str]:
        query = (
            tables.sep24_transactions.select()
            .where(tables.sep24_transactions.c.kind == "withdrawal")
            .where(tables.sep24_transactions.c.withdraw_anchor_account == account)
            .where(tables.sep24_transactions.c.status == "completed")
            .order_by(tables.sep24_transactions.c.started_at.desc())
        )
        completed_transactions = []
        row = await self.database.fetch_one(query)
        if row:
            return Sep24Transaction.from_db_row(row, assets=self.assets).paging_token
        return None

    async def log_transaction_message(
        self, transaction: Sep24Transaction, message: str
    ):
        query = tables.sep24_transaction_logs.insert()
        await self.database.execute(
            query=query,
            values={
                "timestamp": datetime.now(timezone.utc),
                "transaction_id": str(transaction.id),
                "message": message,
            },
        )

    # TODO mark below method as abstractmethod

    @overrides
    async def send_withdrawal(self, withdrawal: fawaris.Sep24Transaction) -> None:
        print("sending withdrawal to user")

    @overrides
    async def is_withdrawal_complete(
        self, withdrawal: fawaris.Sep24Transaction
    ) -> bool:
        print("withdrawal not complete")
        return False

    @overrides
    async def is_deposit_received(self, deposit: fawaris.Sep24Transaction) -> bool:
        print("deposit received")
        return True

    @overrides
    async def http_get_info(
        self, request: fawaris.Sep24InfoRequest
    ) -> fawaris.Sep24InfoResponse:
        return {
            "deposit": {
                "USDC": fawaris.Sep24InfoResponseDeposit(
                    enabled=True,
                ),
            },
            "withdraw": {
                "USDC": fawaris.Sep24InfoResponseWithdraw(
                    enabled=True,
                ),
            },
        }




import typer
import asyncio
import sys
import signal
import time
import logging
from databases import Database

from fawaris_fastapi import settings
from fawaris_fastapi import Sep24

terminate = False

logger = logging.getLogger(__name__)
database = Database(settings.ASYNC_DATABASE_URL)
sep24 = Sep24.new_instance(database)

def exit_gracefully(*_):  # pragma: no cover
    print("Exiting task_all...")
    module = sys.modules[__name__]
    module.terminate = True

async def task_all():
    try:
        await sep24.task_all()
    except Exception as e:
        logger.exception(e)

def main(loop: bool = False, interval: int = 1):
    def sleep(seconds):  # pragma: no cover
        for _ in range(seconds):
            if terminate:
                break
            time.sleep(interval)

    if loop:
        while True:
            if terminate:
                break
            asyncio.run(task_all())
            sleep(interval)
    else:
        asyncio.run(task_all())


if __name__ == "__main__":
    signal.signal(signal.SIGINT, exit_gracefully)
    signal.signal(signal.SIGTERM, exit_gracefully)
    typer.run(main)



from fastapi import Request
import fawaris

from fawaris_fastapi.utils import detect_and_get_request_data, validate_request_data
from fawaris_fastapi import settings

class Sep10(fawaris.Sep10):
    @classmethod
    def new_instance(cls):
        return cls(
            host_url=settings.HOST_URL,
            home_domains=settings.HOME_DOMAINS,
            horizon_url=settings.HORIZON_URL,
            network_passphrase=settings.NETWORK_PASSPHRASE,
            signing_secret=settings.SIGNING_SECRET,
            jwt_secret=settings.JWT_SECRET,
        )

    def register_routes(self, app):
        @app.get("/auth")
        async def http_get(request: Request):
            params = dict(request.query_params)
            if not params.get("home_domain"):
                params["home_domain"] = settings.HOME_DOMAINS[0]
            req = validate_request_data(params, fawaris.Sep10GetRequest)
            return await self.http_get(req)

        @app.post("/auth")
        async def http_post(request: Request):
            data = await detect_and_get_request_data(request, allowed_content_types=[
                "application/x-www-form-urlencoded", "application/json"
            ])
            req = validate_request_data(data, fawaris.Sep10PostRequest)
            return await self.http_post(req)



from pydantic.error_wrappers import ValidationError

class RequestValidationError(ValidationError):
    def __init__(self, validation_error: ValidationError):
        super().__init__(validation_error.raw_errors, validation_error.model)



import sqlalchemy
import typer
from fawaris_fastapi.tables import metadata
from fawaris_fastapi import settings

def main():
    engine = sqlalchemy.create_engine(
        settings.DATABASE_URL, connect_args={"check_same_thread": False}
    )
    metadata.create_all(engine)


if __name__ == "__main__":
    typer.run(main)



import sqlalchemy
from sqlalchemy.dialects import sqlite

metadata = sqlalchemy.MetaData()

sep24_transactions = sqlalchemy.Table(
    "sep24_transactions",
    metadata,

    # extra fields, specific to this implementation
    sqlalchemy.Column("asset_code", sqlalchemy.Text),
    sqlalchemy.Column("asset_issuer", sqlalchemy.Text),
    sqlalchemy.Column("paging_token", sqlalchemy.Text),
    sqlalchemy.Column("claimable_balance_supported", sqlalchemy.Boolean),
    sqlalchemy.Column("stellar_transaction_response", sqlalchemy.JSON),

    # fields 1-to-1 to fawaris.Sep24Transaction
    sqlalchemy.Column("id", sqlalchemy.Text, primary_key=True),
    sqlalchemy.Column("kind", sqlalchemy.Text),
    sqlalchemy.Column("status", sqlalchemy.Text),
    sqlalchemy.Column("status_eta", sqlalchemy.Integer),
    sqlalchemy.Column("more_info_url", sqlalchemy.Text),
    sqlalchemy.Column("amount_in", sqlalchemy.Text),
    sqlalchemy.Column("amount_in_asset", sqlalchemy.Text),
    sqlalchemy.Column("amount_out", sqlalchemy.Text),
    sqlalchemy.Column("amount_out_asset", sqlalchemy.Text),
    sqlalchemy.Column("amount_fee", sqlalchemy.Text),
    sqlalchemy.Column("amount_fee_asset", sqlalchemy.Text),
    sqlalchemy.Column("from_address", sqlalchemy.Text),
    sqlalchemy.Column("to_address", sqlalchemy.Text),
    sqlalchemy.Column("external_extra", sqlalchemy.Text),
    sqlalchemy.Column("external_extra_text", sqlalchemy.Text),
    sqlalchemy.Column("deposit_memo", sqlalchemy.Text),
    sqlalchemy.Column("deposit_memo_type", sqlalchemy.Text),
    sqlalchemy.Column("withdraw_anchor_account", sqlalchemy.Text),
    sqlalchemy.Column("withdraw_memo", sqlalchemy.Text),
    sqlalchemy.Column("withdraw_memo_type", sqlalchemy.Text),
    sqlalchemy.Column("started_at", sqlalchemy.DateTime),
    sqlalchemy.Column("completed_at", sqlalchemy.DateTime),
    sqlalchemy.Column("stellar_transaction_id", sqlalchemy.Text),
    sqlalchemy.Column("external_transaction_id", sqlalchemy.Text),
    sqlalchemy.Column("message", sqlalchemy.Text),
    sqlalchemy.Column("refunds", sqlalchemy.JSON),
    sqlalchemy.Column("required_info_message", sqlalchemy.Text),
    sqlalchemy.Column("required_info_updates", sqlalchemy.JSON),
    sqlalchemy.Column("claimable_balance_id", sqlalchemy.Text),
)

sep24_transaction_logs = sqlalchemy.Table(
    "sep24_transaction_logs",
    metadata,
    sqlalchemy.Column("id", sqlalchemy.Integer, primary_key=True, autoincrement=True),
    sqlalchemy.Column("timestamp", sqlalchemy.DateTime),
    sqlalchemy.Column("transaction_id", sqlalchemy.Text),
    sqlalchemy.Column("message", sqlalchemy.Text),
)



from fastapi import FastAPI
from pydantic.error_wrappers import ValidationError
from fastapi.responses import JSONResponse
import fawaris

from fawaris_fastapi.exceptions import RequestValidationError

def register_exception_handlers(app: FastAPI):
    @app.exception_handler(RequestValidationError)
    async def value_error_handler(request, exc):
        msg = ""
        for error in exc.errors():
            if error["type"] == "value_error.missing":
                try:
                    field = error["loc"][0]
                    msg += f"Missing value for '{field}'"
                except (KeyError, IndexError):
                    msg += str(error) + ". "
            else:
                msg += str(error) + ". "
        return JSONResponse({"error": msg}, status_code=400)

    @app.exception_handler(fawaris.Sep10InvalidToken)
    async def sep10_invalid_token_handler(request, exc):
        return JSONResponse({"type": "authentication_required"}, status_code=403)




import asyncio
import codecs
import time
from typing import Optional, Union, Dict
from typing_extensions import Literal
from fastapi import Request
from stellar_sdk import ServerAsync, Keypair, TransactionBuilder, Asset, Claimant
from stellar_sdk.xdr import TransactionResult, OperationType
from pydantic.error_wrappers import ValidationError
from stellar_sdk.client.aiohttp_client import AiohttpClient
from stellar_sdk import TextMemo, IdMemo, HashMemo
import fawaris

from fawaris_fastapi.exceptions import RequestValidationError

class AuthenticatedRequest(Request):
    token = None

def authenticate(jwt_secret: str):
    def wrapper(request: AuthenticatedRequest):
        # TODO handle invalid authorization format
        encoded_jwt = request.headers.get("authorization").split(" ")[1]
        request.token = fawaris.Sep10Token(encoded_jwt, jwt_secret)
        return request
    return wrapper

def validate_request_data(data, model_class):
    try:
        return model_class(**data)
    except ValidationError as e:
        raise RequestValidationError(e)


async def detect_and_get_request_data(
    request: Request,
    allowed_content_types=[
        "multipart/form-data",
        "application/x-www-form-urlencoded",
        "application/json",
    ],
):
    content_type = request.headers.get("content-type")
    allowed = False
    for allowed_content_type in allowed_content_types:
        if allowed_content_type in content_type:
            allowed = True
    if not allowed:
        raise ValueError("Header 'Content-Type' has an invalid value")
    if "multipart/form-data" in content_type:
        data = await request.form()
    elif "application/x-www-form-urlencoded" in content_type:
        data = await request.form()
    elif "application/json" in content_type:
        data = await request.json()
    return data

async def stellar_send_payment(
    source_secret: str,
    horizon_url: str,
    network_passphrase: str,
    destination_account: str,
    amount: str,
    asset_code: str,
    asset_issuer: Optional[str] = None,
    base_fee: Optional[int] = None,
    memo: Optional[Union[str, bytes, int]] = None,
    memo_type: Optional[Literal["text", "hash", "id"]] = None,
    timeout: Optional[int] = 30,
):
    if memo is not None and memo_type is None:
        raise ValueError("'memo_type' is required if 'memo' is set")
    source_kp = Keypair.from_secret(source_secret)
    async with ServerAsync(
        horizon_url=horizon_url, client=AiohttpClient()
    ) as server:
        coroutines = [server.load_account(source_kp.public_key)]
        if base_fee is None:
            coroutines.append(server.fetch_base_fee())
        results = await asyncio.gather(*coroutines)
        source_account = results[0]
        if base_fee is None:
            base_fee = results[1]
        builder = TransactionBuilder(
            source_account=source_account,
            network_passphrase=network_passphrase,
            base_fee=base_fee,
        )
        now = int(time.time())
        builder.add_time_bounds(now - 600, now + timeout)
        asset = Asset(
            code=asset_code, issuer=asset_issuer
        )
        builder.append_payment_op(
            destination=destination_account,
            asset=asset,
            amount=amount,
            source=source_kp.public_key,
        )
        if memo is not None:
            builder.add_memo(make_memo(memo, memo_type))
        envelope = builder.build()
        envelope.sign(source_kp)
        return await server.submit_transaction(envelope)

async def stellar_create_claimable_balance(
    source_secret: str,
    horizon_url: str,
    network_passphrase: str,
    destination_account: str,
    amount: str,
    asset_code: str,
    asset_issuer: Optional[str] = None,
    base_fee: Optional[int] = None,
    memo: Optional[Union[str, bytes, int]] = None,
    memo_type: Optional[Literal["text", "hash", "id"]] = None,
    timeout: Optional[int] = 30,
):
    if memo is not None and memo_type is None:
        raise ValueError("'memo_type' is required if 'memo' is set")
    source_kp = Keypair.from_secret(source_secret)
    async with ServerAsync(
        horizon_url=horizon_url, client=AiohttpClient()
    ) as server:
        coroutines = [server.load_account(source_kp.public_key)]
        if base_fee is None:
            coroutines.append(server.fetch_base_fee())
        results = await asyncio.gather(*coroutines)
        source_account = results[0]
        if base_fee is None:
            base_fee = results[1]
        builder = TransactionBuilder(
            source_account=source_account,
            network_passphrase=network_passphrase,
            base_fee=base_fee,
        )
        now = int(time.time())
        builder.add_time_bounds(now - 600, now + timeout)
        claimant = Claimant(destination=destination_account)
        asset = Asset(
            code=asset_code, issuer=asset_issuer
        )
        builder.append_create_claimable_balance_op(
            claimants=[claimant],
            asset=asset,
            amount=amount,
            source=source_kp.public_key,
        )
        if memo is not None:
            builder.add_memo(make_memo(memo, memo_type))
        envelope = builder.build()
        envelope.sign(source_kp)
        return await server.submit_transaction(envelope)

def make_memo(memo: str, memo_type: str) -> Optional[Union[TextMemo, HashMemo, IdMemo]]:
    if not (memo or memo_type):
        return None
    if memo_type == "id":
        return IdMemo(int(memo))
    elif memo_type == "hash":
        return HashMemo(memo_base64_to_hex(memo))
    elif memo_type == "text":
        return TextMemo(memo)
    else:
        raise ValueError(f"Invalid memo_type: {memo_type}")

def memo_base64_to_hex(memo):
    return (
        codecs.encode(codecs.decode(memo.encode(), "base64"), "hex").decode("utf-8")
    ).strip()

async def fetch_account(horizon_url: str, account: str) -> dict:
    async with ServerAsync(
        horizon_url=horizon_url, client=AiohttpClient()
    ) as server:
        return await server.accounts().account_id(account).call()

def is_pending_trust(account_dict: dict, asset_code: str, asset_issuer: str):
    pending_trust = True
    for balance in account_dict["balances"]:
        if balance.get("asset_type") in ["native", "liquidity_pool_shares"]:
            continue
        if (
            asset_code == balance["asset_code"]
            and asset_issuer == balance["asset_issuer"]
        ):
            pending_trust = False
            break
    return pending_trust

def get_claimable_balance_id(horizon_response: dict) -> Optional[str]:
    result_xdr = horizon_response["result_xdr"]
    balance_id_hex = None
    for op_result in TransactionResult.from_xdr(result_xdr).result.results:
        if op_result.tr.type == OperationType.CREATE_CLAIMABLE_BALANCE:
            balance_id_hex = (
                op_result.tr.create_claimable_balance_result.balance_id.to_xdr_bytes().hex()
            )
    return balance_id_hex



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['fawaris_fastapi', 'fawaris_fastapi.scripts']

package_data = \
{'': ['*']}

install_requires = \
['aiosqlite>=0.17.0,<0.18.0',
 'databases>=0.5.5,<0.6.0',
 'fastapi>=0.75.0,<0.76.0',
 'fawaris>=0.1.36,<0.2.0',
 'overrides>=6.1.0,<7.0.0',
 'python-multipart>=0.0.5,<0.0.6',
 'typer>=0.4.0,<0.5.0',
 'uvicorn>=0.17.6,<0.18.0']

setup_kwargs = {
    'name': 'fawaris-fastapi',
    'version': '0.1.0',
    'description': 'Stellar Anchor implementation using FastAPI',
    'long_description': None,
    'author': 'Yuri Escalianti',
    'author_email': 'yuriescl@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import sqlalchemy
from databases import Database
from fastapi import FastAPI, Form, Request
from fastapi.responses import JSONResponse
from stellar_sdk import Network, Keypair
from pydantic.error_wrappers import ValidationError
import fawaris

from fawaris_fastapi import Sep10, Sep24, metadata, settings
from fawaris_fastapi.exception_handlers import (
    register_exception_handlers
)

app = FastAPI()

database = Database(settings.ASYNC_DATABASE_URL)

sep10 = Sep10.new_instance()
sep24 = Sep24.new_instance(database)

register_exception_handlers(app)
sep10.register_routes(app)
sep24.register_routes(app)

@app.on_event("startup")
async def startup():
    await database.connect()

@app.on_event("shutdown")
async def shutdown():
    await database.disconnect()
